@@ -52,9 +52,10 @@
 
 #define le8dec(x) (*x)
 
-#ifndef le16dec  /* FreeBSD standard functions not available */
+#ifndef le16dec  /* FreeBSD marshalling functions not available */
 
 #ifdef OS_DARWIN
+/* Darwin platform */
 #include <libkern/OSByteOrder.h>
 
 #define le16dec(ptr) OSReadLittleInt16(ptr, 0)
@@ -73,10 +74,70 @@
 #define be32enc(ptr, n) OSWriteBigInt32(ptr, 0, n)
 #define be64enc(ptr, n) OSWriteBigInt64(ptr, 0, n)
 
-#else  /* OS_DARWIN */
+#else
 
+/* Non-Darwin platform */
 /* Emulate with memcpy and endianness conversion functions */
 
+/* Define BSD conversion functions if undefined */
+#ifndef le16toh
+
+/* Define portable byte swapping if undefined */
+/* WARNING: will not work on mixed endian machines! */
+/* WARNING: these functions must be defined by the system in the above case */
+#ifndef bswap_16
+#define bswap_16(value) \
+  (((uint16_t) ((value) & 0xff) << 8) | ((uint16_t) (value) >> 8))
+#endif
+
+#ifndef bswap_32
+#define bswap_32(value) \
+  (((uint32_t) bswap_16(((value) & 0xffff)) << 16) | \
+    (uint32_t) bswap_16(((value) >> 16)))
+#endif
+
+#ifndef bswap_64
+#define bswap_64(value) \
+  (((uint64_t) bswap_32(((value) & 0xffffffff)) << 32) | \
+    (uint64_t) bswap_32(((value) >> 32)))
+#endif
+
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+/* Little-endian architecture */
+#define htobe16(x) bswap_16 (x)
+#define htole16(x) (x)
+#define be16toh(x) bswap_16 (x)
+#define le16toh(x) (x)
+
+#define htobe32(x) bswap_32 (x)
+#define htole32(x) (x)
+#define be32toh(x) bswap_32 (x)
+#define le32toh(x) (x)
+
+#define htobe64(x) bswap_64 (x)
+#define htole64(x) (x)
+#define be64toh(x) bswap_64 (x)
+#define le64toh(x) (x)
+#else
+/* Big-endian architecture */
+#define htobe16(x) (x)
+#define htole16(x) bswap_16 (x)
+#define be16toh(x) (x)
+#define le16toh(x) bswap_16 (x)
+
+#define htobe32(x) (x)
+#define htole32(x) bswap_32 (x)
+#define be32toh(x) (x)
+#define le32toh(x) bswap_32 (x)
+
+#define htobe64(x) (x)
+#define htole64(x) bswap_64 (x)
+#define be64toh(x) (x)
+#define le64toh(x) bswap_64 (x)
+#endif  /* byte order */
+
+#endif  /* BSD conversion functions */
+
 #define le16dec(ptr) \
   (__extension__ ({ uint16_t __n; memcpy(&__n, ptr, 2); le16toh(__n); }))
 #define le32dec(ptr) \
@@ -107,7 +168,7 @@
 
 #endif  /* OS_DARWIN */
 
-#endif  /* le16dec */
+#endif  /* FreeBSD marshalling functions */
 
 
 /* Bin-prot integer codes */
