@@ -1,28 +1,26 @@
-(*pp cpp $ARCH_FLAGS *)
-
-(* File: read_ml.ml
-
-    Copyright (C) 2007-
-
-      Jane Street Holding, LLC
-      Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
-
-   This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
-
-   This library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this library; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*)
+(******************************************************************************
+ *                             Bin-prot                                       *
+ *                                                                            *
+ * Copyright (C) 2008- Jane Street Holding, LLC                               *
+ *    Contact: opensource@janestreet.com                                      *
+ *    WWW: http://www.janestreet.com/ocaml                                    *
+ *    Author: Markus Mottl                                                    *
+ *                                                                            *
+ * This library is free software; you can redistribute it and/or              *
+ * modify it under the terms of the GNU Lesser General Public                 *
+ * License as published by the Free Software Foundation; either               *
+ * version 2 of the License, or (at your option) any later version.           *
+ *                                                                            *
+ * This library is distributed in the hope that it will be useful,            *
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
+ * Lesser General Public License for more details.                            *
+ *                                                                            *
+ * You should have received a copy of the GNU Lesser General Public           *
+ * License along with this library; if not, write to the Free Software        *
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
+ *                                                                            *
+ ******************************************************************************)
 
 (* Read_ml: reading values from the binary protocol using (mostly) OCaml. *)
 
@@ -44,6 +42,10 @@ type ('a, 'b) reader1 = 'a reader -> 'b reader
 type ('a, 'b, 'c) reader2 = 'a reader -> ('b, 'c) reader1
 type ('a, 'b, 'c, 'd) reader3 = 'a reader -> ('b, 'c, 'd) reader2
 
+let rewrite_c_error f buf ~pos_ref =
+  try f buf ~pos_ref
+  with Unsafe_read_c.Error read_err -> raise_read_error read_err !pos_ref
+
 let bin_read_unit (buf : buf) ~pos_ref =
   let pos = safe_get_pos buf pos_ref in
   if buf.{pos} = '\000' then pos_ref := pos + 1
@@ -216,8 +218,7 @@ let bin_read_nat0 buf ~pos_ref =
 
 let bin_read_string buf ~pos_ref =
   let start_pos = !pos_ref in
-  let plen = bin_read_nat0 buf ~pos_ref in
-  let len = Nat0.to_int plen in
+  let len = (bin_read_nat0 buf ~pos_ref :> int) in
   if len > Sys.max_string_length then
     raise_read_error ReadError.String_too_long start_pos;
   let pos = !pos_ref in
@@ -251,6 +252,8 @@ let bin_read_int buf ~pos_ref =
 external bin_read_float :
   buf -> pos_ref : int ref -> float = "ml_read_float_stub"
 
+let bin_read_float buf ~pos_ref = rewrite_c_error bin_read_float buf ~pos_ref
+
 let read_int32_aux buf ~pos_ref =
 #ifdef ARCH_SIXTYFOUR
   let n = safe_bin_read_int32 buf ~pos_ref in
@@ -390,21 +393,19 @@ let rec bin_read_n_rev_list bin_read_el buf ~pos_ref len =
   loop len []
 
 let bin_read_list bin_read_el buf ~pos_ref =
-  let plen = bin_read_nat0 buf ~pos_ref in
-  let len = Nat0.to_int plen in
+  let len = (bin_read_nat0 buf ~pos_ref :> int) in
   let rev_lst = bin_read_n_rev_list bin_read_el buf ~pos_ref len in
   List.rev rev_lst
 
 #ifndef ARCH_SIXTYFOUR
-let dummy_float_buf = Array1.create char c_layout 8
+let dummy_float_buf = create_buf 8
 let () = ignore (Write_ml.bin_write_float dummy_float_buf ~pos:0 3.1)
 let max_array_length_2 = Sys.max_array_length / 2
 #endif
 
 let bin_read_array bin_read_el buf ~pos_ref =
   let start_pos = !pos_ref in
-  let plen = bin_read_nat0 buf ~pos_ref in
-  let len = Nat0.to_int plen in
+  let len = (bin_read_nat0 buf ~pos_ref :> int) in
   if len = 0 then [||]
   else (
 #ifdef ARCH_SIXTYFOUR
@@ -438,8 +439,7 @@ let bin_read_array bin_read_el buf ~pos_ref =
     res)
 
 let bin_read_hashtbl bin_read_key bin_read_val buf ~pos_ref =
-  let plen = bin_read_nat0 buf ~pos_ref in
-  let len = Nat0.to_int plen in
+  let len = (bin_read_nat0 buf ~pos_ref :> int) in
   let htbl = Hashtbl.create len in
   let read_kv_pair = bin_read_pair bin_read_key bin_read_val in
   let els = bin_read_n_rev_list read_kv_pair buf ~pos_ref len in
@@ -448,29 +448,117 @@ let bin_read_hashtbl bin_read_key bin_read_val buf ~pos_ref =
 external bin_read_float32_vec :
   buf -> pos_ref : pos_ref -> vec32 = "ml_read_float32_vec_stub"
 
+let bin_read_float32_vec buf ~pos_ref =
+  rewrite_c_error bin_read_float32_vec buf ~pos_ref
+
 external bin_read_float64_vec :
   buf -> pos_ref : pos_ref -> vec64 = "ml_read_float64_vec_stub"
 
+let bin_read_float64_vec buf ~pos_ref =
+  rewrite_c_error bin_read_float64_vec buf ~pos_ref
+
 external bin_read_vec :
   buf -> pos_ref : pos_ref -> vec = "ml_read_float64_vec_stub"
 
+let bin_read_vec buf ~pos_ref = rewrite_c_error bin_read_vec buf ~pos_ref
+
 external bin_read_float32_mat :
   buf -> pos_ref : pos_ref -> mat32 = "ml_read_float32_mat_stub"
 
+let bin_read_float32_mat buf ~pos_ref =
+  rewrite_c_error bin_read_float32_mat buf ~pos_ref
+
 external bin_read_float64_mat :
   buf -> pos_ref : pos_ref -> mat64 = "ml_read_float64_mat_stub"
 
+let bin_read_float64_mat buf ~pos_ref =
+  rewrite_c_error bin_read_float64_mat buf ~pos_ref
+
 external bin_read_mat :
   buf -> pos_ref : pos_ref -> mat = "ml_read_float64_mat_stub"
 
+let bin_read_mat buf ~pos_ref = rewrite_c_error bin_read_mat buf ~pos_ref
+
 external bin_read_bigstring :
   buf -> pos_ref : pos_ref -> buf = "ml_read_bigstring_stub"
 
+let bin_read_bigstring buf ~pos_ref =
+  rewrite_c_error bin_read_bigstring buf ~pos_ref
+
 external bin_read_float_array :
   buf -> pos_ref : pos_ref -> float array = "ml_read_float_array_stub"
 
+let bin_read_float_array buf ~pos_ref =
+  rewrite_c_error bin_read_float_array buf ~pos_ref
+
 external bin_read_variant_int :
   buf -> pos_ref : pos_ref -> int = "ml_read_variant_tag_stub"
 
+let bin_read_variant_int buf ~pos_ref =
+  rewrite_c_error bin_read_variant_int buf ~pos_ref
+
 external bin_read_variant_tag :
   buf -> pos_ref : pos_ref -> [> ] = "ml_read_variant_tag_stub"
+
+let bin_read_variant_tag buf ~pos_ref =
+  rewrite_c_error bin_read_variant_tag buf ~pos_ref
+
+external bin_read_int_8bit :
+  buf -> pos_ref : pos_ref -> int = "ml_read_char_stub"
+
+let bin_read_int_8bit buf ~pos_ref =
+  rewrite_c_error bin_read_int_8bit buf ~pos_ref
+
+external bin_read_int_16bit :
+  buf -> pos_ref : pos_ref -> int = "ml_read_int_16bit_stub"
+
+let bin_read_int_16bit buf ~pos_ref =
+  rewrite_c_error bin_read_int_16bit buf ~pos_ref
+
+external bin_read_int_32bit :
+  buf -> pos_ref : pos_ref -> int = "ml_read_int_32bit_stub"
+
+let bin_read_int_32bit buf ~pos_ref =
+  rewrite_c_error bin_read_int_32bit buf ~pos_ref
+
+external bin_read_int_64bit :
+  buf -> pos_ref : pos_ref -> int = "ml_read_int_64bit_stub"
+
+let bin_read_int_64bit buf ~pos_ref =
+  rewrite_c_error bin_read_int_64bit buf ~pos_ref
+
+external bin_read_int64_bits :
+  buf -> pos_ref : pos_ref -> int64 = "ml_read_int64_bits_stub"
+
+let bin_read_int64_bits buf ~pos_ref =
+  rewrite_c_error bin_read_int64_bits buf ~pos_ref
+
+external bin_read_network16_int :
+  buf -> pos_ref : pos_ref -> int = "ml_read_network16_int_stub"
+
+let bin_read_network16_int buf ~pos_ref =
+  rewrite_c_error bin_read_network16_int buf ~pos_ref
+
+external bin_read_network32_int :
+  buf -> pos_ref : pos_ref -> int = "ml_read_network32_int_stub"
+
+let bin_read_network32_int buf ~pos_ref =
+  rewrite_c_error bin_read_network32_int buf ~pos_ref
+
+external bin_read_network32_int32 :
+  buf -> pos_ref : pos_ref -> int32 = "ml_read_network32_int32_stub"
+
+let bin_read_network32_int32 buf ~pos_ref =
+  rewrite_c_error bin_read_network32_int32 buf ~pos_ref
+
+external bin_read_network64_int :
+  buf -> pos_ref : pos_ref -> int = "ml_read_network64_int_stub"
+
+let bin_read_network64_int buf ~pos_ref =
+  rewrite_c_error bin_read_network64_int buf ~pos_ref
+
+external bin_read_network64_int64 :
+  buf -> pos_ref : pos_ref -> int64 = "ml_read_network64_int64_stub"
+
+let bin_read_network64_int64 buf ~pos_ref =
+  rewrite_c_error bin_read_network64_int64 buf ~pos_ref
