@@ -1,26 +1,26 @@
-/* File: read_stubs.c
-
-    Copyright (C) 2007-
-
-      Jane Street Holding, LLC
-      Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
-
-   This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
-
-   This library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this library; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*/
+/******************************************************************************
+ *                             Bin-prot                                       *
+ *                                                                            *
+ * Copyright (C) 2008- Jane Street Holding, LLC                               *
+ *    Contact: opensource@janestreet.com                                      *
+ *    WWW: http://www.janestreet.com/ocaml                                    *
+ *    Author: Markus Mottl                                                    *
+ *                                                                            *
+ * This library is free software; you can redistribute it and/or              *
+ * modify it under the terms of the GNU Lesser General Public                 *
+ * License as published by the Free Software Foundation; either               *
+ * version 2 of the License, or (at your option) any later version.           *
+ *                                                                            *
+ * This library is distributed in the hope that it will be useful,            *
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
+ * Lesser General Public License for more details.                            *
+ *                                                                            *
+ * You should have received a copy of the GNU Lesser General Public           *
+ * License along with this library; if not, write to the Free Software        *
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
+ *                                                                            *
+ ******************************************************************************/
 
 /* Stubs for reading basic values in the binary protocol */
 
@@ -43,7 +43,6 @@ CAMLprim value bin_prot_read_ml_init_stub(value __unused v_unit)
   return Val_unit;
 }
 
-
 /* Raising errors */
 
 /* Constant values come from the order of variants in common.ml */
@@ -83,12 +82,6 @@ static inline void raise_Read_error(int loc, unsigned long pos)
 
 /* Utility macros */
 
-#define MK_DO_READ(SIZE, TYPE) \
-  static inline __pure TYPE do_read_int##SIZE(char **sptr_ptr) \
-  { \
-    return *(TYPE *) *sptr_ptr; \
-  }
-
 #define MK_GEN_SAFE_READ(NAME, SIZE, TYPE, LEN, CHECK) \
   static inline TYPE safe_read_##NAME##SIZE(char **sptr_ptr, char *eptr) \
   { \
@@ -97,21 +90,20 @@ static inline void raise_Read_error(int loc, unsigned long pos)
     TYPE n; \
     if (unlikely(next > eptr)) \
       caml_raise_constant(*v_bin_prot_exc_Buffer_short); \
-    n = do_read_int##SIZE(sptr_ptr); \
+    n = le##SIZE##dec(sptr); \
     CHECK \
     *sptr_ptr = next; \
     return n; \
   }
 
 #define MK_SAFE_READ(SIZE, TYPE, LEN, CHECK) \
-  MK_DO_READ(SIZE, TYPE) \
   MK_GEN_SAFE_READ(int, SIZE, TYPE, LEN, CHECK)
 
 #define MK_GEN_SAFE_NAT0_READ(PREF, SIZE, TYPE, LEN, CHECK) \
   MK_GEN_SAFE_READ(PREF##nat0_, SIZE, unsigned TYPE, LEN, CHECK)
 
 #define MK_SAFE_NAT0_READ(SIZE, TYPE, LEN, CHECK) \
-  MK_GEN_SAFE_NAT0_READ(, SIZE, TYPE, LEN, CHECK)
+  MK_GEN_SAFE_READ(nat0_, SIZE, unsigned TYPE, LEN, CHECK)
 
 #define MK_ML_READER(NAME) \
   CAMLprim value ml_read_##NAME##_stub(value v_buf, value v_pos_ref) \
@@ -133,26 +125,23 @@ static inline void raise_Read_error(int loc, unsigned long pos)
 
 /* Reading OCaml integers */
 
-MK_DO_READ(8, char)
-
 MK_GEN_SAFE_READ(neg_int, 8, char, 1,
   if (unlikely(n >= 0)) {
     *sptr_ptr = sptr - 1;
     raise_Error(READ_ERROR_NEG_INT8);
   })
 
-MK_SAFE_READ(16, short, 2,)
+MK_SAFE_READ(16, short, 2, {})
 
 #ifdef ARCH_SIXTYFOUR
-  MK_SAFE_READ(32, int, 4,)
+  MK_SAFE_READ(32, int, 4, {})
 #else
-  MK_DO_READ(32, int)
   MK_GEN_SAFE_READ(int, 32, int, 4,
     if (unlikely(n < -0x40000000l || n > 0x3FFFFFFFl)) {
       *sptr_ptr = sptr - 1;
       raise_Error(READ_ERROR_INT_OVERFLOW);
     })
-  MK_GEN_SAFE_READ(nocheck_int, 32, int, 4,)
+  MK_GEN_SAFE_READ(nocheck_int, 32, int, 4, {})
 #endif
 
 #ifdef ARCH_SIXTYFOUR
@@ -161,7 +150,7 @@ MK_SAFE_READ(16, short, 2,)
       *sptr_ptr = sptr - 1;
       raise_Error(READ_ERROR_INT_OVERFLOW);
     })
-  MK_GEN_SAFE_READ(nocheck_int, 64, long, 8,)
+  MK_GEN_SAFE_READ(nocheck_int, 64, long, 8, {})
 #endif
 
 static inline long read_int(char **sptr_ptr, char *eptr)
@@ -190,17 +179,17 @@ CAMLprim value read_int_stub(char **sptr_ptr, char *eptr)
 
 /* Non-negative OCaml integers */
 
-MK_SAFE_NAT0_READ(16, short, 2,)
+MK_SAFE_NAT0_READ(16, short, 2, {})
 
 #ifdef ARCH_SIXTYFOUR
-  MK_SAFE_NAT0_READ(32, int, 4,)
+  MK_SAFE_NAT0_READ(32, int, 4, {})
 #else
   MK_SAFE_NAT0_READ(32, int, 4,
     if (unlikely(n > 0x3FFFFFFFl)) {
       *sptr_ptr = sptr - 1;
       raise_Error(READ_ERROR_NAT0_OVERFLOW);
     })
-  MK_GEN_SAFE_NAT0_READ(nocheck, 32, int, 4,)
+  MK_GEN_SAFE_NAT0_READ(nocheck, 32, int, 4, {})
 #endif
 
 #ifdef ARCH_SIXTYFOUR
@@ -209,7 +198,6 @@ MK_SAFE_NAT0_READ(16, short, 2,)
       *sptr_ptr = sptr - 1;
       raise_Error(READ_ERROR_NAT0_OVERFLOW);
     })
-  MK_GEN_SAFE_NAT0_READ(nocheck, 64, long, 8,)
 #endif
 
 static inline unsigned long read_nat0(char **sptr_ptr, char *eptr)
@@ -265,12 +253,6 @@ CAMLprim value read_int32_stub(char **sptr_ptr, char *eptr)
 
 /* Reading 64bit integers */
 
-#ifdef ARCH_INT64_TYPE
-#include "int64_native.h"
-#else
-#include "int64_emul.h"
-#endif
-
 static inline int64 read_int64(char **sptr_ptr, char *eptr)
 {
   char *sptr = *sptr_ptr;
@@ -278,14 +260,15 @@ static inline int64 read_int64(char **sptr_ptr, char *eptr)
   if (unlikely(sptr >= eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
   code = *sptr;
   *sptr_ptr = sptr + 1;
-  if (likely(code >= 0))
 #ifdef ARCH_SIXTYFOUR
+  if (likely(code >= 0))
     return code;
   if (likely(code == CODE_INT16)) return safe_read_int16(sptr_ptr, eptr);
   if (likely(code == CODE_NEG_INT8)) return safe_read_neg_int8(sptr_ptr, eptr);
   if (likely(code == CODE_INT32)) return safe_read_int32(sptr_ptr, eptr);
   if (likely(code == CODE_INT64)) return safe_read_nocheck_int64(sptr_ptr, eptr);
 #else
+  if (likely(code >= 0))
     return I64_literal(0, code);
   if (likely(code == CODE_INT16))
     return I64_of_int32(safe_read_int16(sptr_ptr, eptr));
@@ -336,15 +319,6 @@ CAMLprim value read_nativeint_stub(char **sptr_ptr, char *eptr)
 }
 
 
-/* Reading unsigned short words */
-
-CAMLprim value read_uint16_stub(char **sptr_ptr, char *eptr)
-{
-  unsigned short res = safe_read_nat0_16(sptr_ptr, eptr);
-  return Val_int(res);
-}
-
-
 /* Reading unit value */
 
 CAMLprim value read_unit_stub(char **sptr_ptr, char *eptr)
@@ -382,12 +356,13 @@ READ_ZERO_OR_ONE(option_bool, READ_ERROR_OPTION_CODE)
 CAMLprim value read_char_stub(char **sptr_ptr, char *eptr)
 {
   char *sptr = *sptr_ptr;
-  int res;
+  unsigned char res;
   if (unlikely(sptr >= eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
-  res = *sptr;
+  res = (unsigned char) *sptr;
   *sptr_ptr = ++sptr;
   return Val_int(res);
 }
+MK_ML_READER(char)
 
 
 /* Reading strings */
@@ -417,9 +392,11 @@ CAMLprim inline value read_float_stub(char **sptr_ptr, char *eptr)
 {
   char *sptr = *sptr_ptr;
   char *next = sptr + sizeof(double);
+  double n;
   if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
   *sptr_ptr = next;
-  return caml_copy_double(*(double *) sptr);
+  memcpy(&n, sptr, sizeof(double));
+  return caml_copy_double(n);
 }
 
 MK_ML_READER(float)
@@ -490,7 +467,7 @@ CAMLprim value read_variant_tag_stub(char **sptr_ptr, char *eptr)
   char *next = sptr + 4;
   int n;
   if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
-  n = *(int *) sptr;
+  n = le32dec(sptr);
   if (likely(Is_long(n))) {
     *sptr_ptr = next;
     return (value) n;
@@ -504,12 +481,12 @@ CAMLprim value ml_read_variant_tag_stub(value v_buf, value v_pos_ref)
   char *start = buf->data;
   long pos = Long_val(Field(v_pos_ref, 0));
   char *sptr = start + pos;
-  int next_pos = pos + 4;
+  unsigned long next_pos = (unsigned long) pos + 4;
   int n;
   if (unlikely(pos < 0)) caml_array_bound_error();
-  if (unlikely(next_pos > *buf->dim))
+  if (unlikely(next_pos > (unsigned long) *buf->dim))
     caml_raise_constant(*v_bin_prot_exc_Buffer_short);
-  n = *(int *) sptr;
+  n = le32dec(sptr);
   if (likely(Is_long(n))) {
     Field(v_pos_ref, 0) = Val_long(next_pos);
     return (value) n;
@@ -518,13 +495,28 @@ CAMLprim value ml_read_variant_tag_stub(value v_buf, value v_pos_ref)
 }
 
 
+/* Reading raw strings */
+
+CAMLprim inline value read_raw_string_stub(
+  char **sptr_ptr, char *eptr, value v_str, value v_pos, value v_len)
+{
+  size_t pos = (size_t) Long_val(v_pos), len = (size_t) Long_val(v_len);
+  char *sptr = *sptr_ptr;
+  char *next = sptr + len;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+  *sptr_ptr = next;
+  memcpy(String_val(v_str) + pos, sptr, len);
+  return Val_unit;
+}
+
+
 /* Reading bigarrays */
 
-#define MK_BA1_READER(NAME, TYPE, TFLAG) \
+#define MK_BA1_READER(NAME, TYPE, TFLAG, TLAYOUT) \
   CAMLprim inline value read_##NAME##_stub(char **sptr_ptr, char *eptr) \
   { \
     unsigned long len = read_nat0(sptr_ptr, eptr); \
-    int tot_size = len * sizeof(TYPE); \
+    unsigned long tot_size = len * sizeof(TYPE); \
     char *sptr = *sptr_ptr; \
     char *next = sptr + tot_size; \
     intnat dim; \
@@ -534,7 +526,7 @@ CAMLprim value ml_read_variant_tag_stub(value v_buf, value v_pos_ref)
     dim = len; \
     v_res = \
       caml_ba_alloc( \
-        CAML_BA_##TFLAG | CAML_BA_FORTRAN_LAYOUT, 1, NULL, &dim); \
+        CAML_BA_##TFLAG | CAML_BA_##TLAYOUT##_LAYOUT, 1, NULL, &dim); \
     *sptr_ptr = next; \
     if (unlikely(tot_size > 65536)) { \
       Begin_roots1(v_res); \
@@ -548,17 +540,17 @@ CAMLprim value ml_read_variant_tag_stub(value v_buf, value v_pos_ref)
   \
   MK_ML_READER(NAME)
 
-MK_BA1_READER(bigstring, char, UINT8)
+MK_BA1_READER(bigstring, char, UINT8, C)
 
 #define MK_VEC_MAT_READERS(NAME, TYPE, TFLAG) \
-  MK_BA1_READER(NAME##_vec, TYPE, TFLAG) \
+  MK_BA1_READER(NAME##_vec, TYPE, TFLAG, FORTRAN) \
   \
   CAMLprim inline value read_##NAME##_mat_stub(char **sptr_ptr, char *eptr) \
   { \
     unsigned long dim1 = read_nat0(sptr_ptr, eptr); \
     unsigned long dim2 = read_nat0(sptr_ptr, eptr); \
     unsigned long size = dim1 * dim2; \
-    int tot_size = size * sizeof(TYPE); \
+    unsigned long tot_size = size * sizeof(TYPE); \
     char *sptr = *sptr_ptr; \
     char *next = sptr + tot_size; \
     intnat dims[2]; \
@@ -585,3 +577,146 @@ MK_BA1_READER(bigstring, char, UINT8)
 
 MK_VEC_MAT_READERS(float32, float, FLOAT32)
 MK_VEC_MAT_READERS(float64, double, FLOAT64)
+
+
+/* Reading bits */
+
+CAMLprim value read_int_16bit_stub(char **sptr_ptr, char *eptr)
+{
+  unsigned short res = safe_read_nat0_16(sptr_ptr, eptr);
+  return Val_int(res);
+}
+MK_ML_READER(int_16bit)
+
+CAMLprim value read_int_32bit_stub(char **sptr_ptr, char *eptr)
+{
+  unsigned int res = safe_read_nat0_32(sptr_ptr, eptr);
+  return Val_int(res);
+}
+MK_ML_READER(int_32bit)
+
+CAMLprim inline value read_int_64bit_stub(char **sptr_ptr, char *eptr)
+{
+  long n;
+#ifndef ARCH_SIXTYFOUR
+  long upper;
+#endif
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 8;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+#ifdef ARCH_SIXTYFOUR
+  n = le64dec(sptr);
+  if (unlikely(n < -0x4000000000000000L || n > 0x3FFFFFFFFFFFFFFFL))
+    raise_Error(READ_ERROR_INT_OVERFLOW);
+#else
+  n = le32dec(sptr);
+  memcpy(&upper, sptr + 4, 4);
+  if (upper == 0l) {
+    if ((unsigned long) n > 0x3FFFFFFFl) raise_Error(READ_ERROR_INT_OVERFLOW);
+  } else if (upper == -1) {
+    if (n < -0x40000000l) raise_Error(READ_ERROR_INT_OVERFLOW);
+  }
+  else raise_Error(READ_ERROR_INT_OVERFLOW);
+#endif
+  *sptr_ptr = next;
+  return Val_long(n);
+}
+MK_ML_READER(int_64bit)
+
+CAMLprim inline value read_int64_bits_stub(char **sptr_ptr, char *eptr)
+{
+  int64 n;
+  value v_res;
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 8;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+#ifdef ARCH_SIXTYFOUR
+  n = le64dec(sptr);
+#else
+  n = I64_literal(le32dec(sptr + 4), le32dec(sptr));
+#endif
+  v_res = caml_copy_int64(n);
+  *sptr_ptr = next;
+  return (value) v_res;
+}
+MK_ML_READER(int64_bits)
+
+CAMLprim inline value read_network16_int_stub(char **sptr_ptr, char *eptr)
+{
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 2;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+  *sptr_ptr = next;
+  return (value) Val_int(be16dec(sptr));
+}
+MK_ML_READER(network16_int)
+
+CAMLprim inline value read_network32_int_stub(char **sptr_ptr, char *eptr)
+{
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 4;
+  int n;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+  n = (int) be32dec(sptr);
+#ifdef ARCH_SIXTYFOUR
+  *sptr_ptr = next;
+  return (value) Val_int((uint32_t) n);
+#else
+  if (unlikely(n < -0x40000000l || n > 0x3FFFFFFFl))
+    raise_Error(READ_ERROR_INT_OVERFLOW);
+  *sptr_ptr = next;
+  return (value) Val_int(n);
+#endif
+}
+MK_ML_READER(network32_int)
+
+CAMLprim inline value read_network32_int32_stub(char **sptr_ptr, char *eptr)
+{
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 4;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+  *sptr_ptr = next;
+  return (value) caml_copy_int32(be32dec(sptr));
+}
+MK_ML_READER(network32_int32)
+
+CAMLprim inline value read_network64_int_stub(char **sptr_ptr, char *eptr)
+{
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 8;
+  long n;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+#ifdef ARCH_SIXTYFOUR
+  n = (long) be64dec(sptr);
+  if (unlikely(n < -0x4000000000000000L || n > 0x3FFFFFFFFFFFFFFFL))
+    raise_Error(READ_ERROR_INT_OVERFLOW);
+#else /* 32bit */
+  /* Read the upper 32 bits first.  They must all be zero, otherwise we
+     consider this an overflow.  On 32bit platforms the integer must
+     fit completely into one word. */
+  memcpy(&n, sptr, 4);
+  if (n != 0) raise_Error(READ_ERROR_INT_OVERFLOW);
+  n = be32dec(sptr + 4);
+  if (unlikely(n < -0x40000000l || n > 0x3FFFFFFFl))
+    raise_Error(READ_ERROR_INT_OVERFLOW);
+#endif
+  *sptr_ptr = next;
+  return (value) Val_long(n);
+}
+MK_ML_READER(network64_int)
+
+CAMLprim inline value read_network64_int64_stub(char **sptr_ptr, char *eptr)
+{
+  char *sptr = *sptr_ptr;
+  char *next = sptr + 8;
+  int64 n;
+  if (unlikely(next > eptr)) caml_raise_constant(*v_bin_prot_exc_Buffer_short);
+  *sptr_ptr = next;
+#ifdef ARCH_SIXTYFOUR
+  n = (long) be64dec(sptr);
+#else /* 32bit */
+  n = I64_literal(be32dec(sptr), be32dec(sptr + 4));
+#endif
+  return (value) caml_copy_int64(n);
+}
+MK_ML_READER(network64_int64)
