@@ -1,6 +1,26 @@
-(*pp camlp4o -I ../lib -I `ocamlfind query type-conv` pa_type_conv.cmo pa_bin_prot.cmo *)
-
-TYPE_CONV_PATH "Bin_prot_test"
+(******************************************************************************
+ *                             Bin-prot                                       *
+ *                                                                            *
+ * Copyright (C) 2008- Jane Street Holding, LLC                               *
+ *    Contact: opensource@janestreet.com                                      *
+ *    WWW: http://www.janestreet.com/ocaml                                    *
+ *    Author: Markus Mottl                                                    *
+ *                                                                            *
+ * This library is free software; you can redistribute it and/or              *
+ * modify it under the terms of the GNU Lesser General Public                 *
+ * License as published by the Free Software Foundation; either               *
+ * version 2 of the License, or (at your option) any later version.           *
+ *                                                                            *
+ * This library is distributed in the hope that it will be useful,            *
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
+ * Lesser General Public License for more details.                            *
+ *                                                                            *
+ * You should have received a copy of the GNU Lesser General Public           *
+ * License along with this library; if not, write to the Free Software        *
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
+ *                                                                            *
+ ******************************************************************************)
 
 open Bigarray
 
@@ -11,6 +31,8 @@ open Bin_prot
 open Common
 open Utils
 open ReadError
+open Type_class
+open Bin_prot.Std
 
 module Bigstring = struct
   type t = buf
@@ -205,6 +227,13 @@ module type Reader_spec = sig
   val bin_read_bigstring : buf reader
   val bin_read_float_array : float array reader
   val bin_read_variant_tag : [> ] reader
+  val bin_read_int_64bit : int reader
+  val bin_read_int64_bits : int64 reader
+  val bin_read_network16_int : int reader
+  val bin_read_network32_int : int reader
+  val bin_read_network32_int32 : int32 reader
+  val bin_read_network64_int : int reader
+  val bin_read_network64_int64 : int64 reader
 end
 
 module type Writer_spec = sig
@@ -237,6 +266,13 @@ module type Writer_spec = sig
   val bin_write_bigstring : buf writer
   val bin_write_float_array : float array writer
   val bin_write_variant_tag : [> ] writer
+  val bin_write_int_64bit : int writer
+  val bin_write_int64_bits : int64 writer
+  val bin_write_network16_int : int writer
+  val bin_write_network32_int : int writer
+  val bin_write_network32_int32 : int32 writer
+  val bin_write_network64_int : int writer
+  val bin_write_network64_int64 : int64 writer
 end
 
 module Make (Spec : SPEC) (Read : Reader_spec) (Write : Writer_spec) = struct
@@ -357,7 +393,6 @@ module Make (Spec : SPEC) (Read : Reader_spec) (Write : Writer_spec) = struct
             let bad_buf = Bigstring.of_string "\132" in
             "Int_code" @? expect_read_error Int_code 0 (fun () ->
               Read.bin_read_int bad_buf ~pos_ref:(ref 0));
-
             if Sys.word_size = 32 then
               let bad_buf = Bigstring.of_string "\253\255\255\255\064" in
               "Int_overflow (positive)" @? expect_read_error Int_overflow 0 (fun () ->
@@ -432,10 +467,13 @@ module Make (Spec : SPEC) (Read : Reader_spec) (Write : Writer_spec) = struct
             let float_tests =
               [
                 mk_float_test 0.;
+                mk_float_test (-0.);
                 mk_float_test (-1.);
                 mk_float_test 1.;
                 mk_float_test infinity;
                 mk_float_test (-.infinity);
+                mk_float_test 1e-310;  (* subnormal *)
+                mk_float_test (-1e-310);  (* subnormal *)
                 mk_float_test 3.141595;
               ]
             in
@@ -845,6 +883,107 @@ module Make (Spec : SPEC) (Read : Reader_spec) (Write : Writer_spec) = struct
               Read.bin_read_variant_tag bad_buf ~pos_ref:(ref 0))
           );
 
+        "int64_bits" >::
+          (fun () ->
+            check_all 8 "int64_bits"
+              Read.bin_read_int64_bits
+              Write.bin_write_int64_bits
+              [
+                (Int64.min_int, "min_int", 8);
+                (Int64.add Int64.min_int Int64.one, "min_int + 1", 8);
+                (Int64.minus_one, "-1", 8);
+                (Int64.zero, "0", 8);
+                (Int64.one, "1", 8);
+                (Int64.sub Int64.max_int Int64.one, "max_int - 1", 8);
+                (Int64.max_int, "max_int", 8);
+              ];
+          );
+
+        "int_64bit" >::
+          (fun () ->
+            check_all 8 "int_64bit"
+              Read.bin_read_int_64bit
+              Write.bin_write_int_64bit
+              [
+                (min_int, "min_int", 8);
+                (min_int + 1, "min_int + 1", 8);
+                (-1, "-1", 8);
+                (0, "0", 8);
+                (1, "1", 8);
+                (max_int - 1, "max_int - 1", 8);
+                (max_int, "max_int", 8);
+              ];
+            let bad_buf_max =
+              bin_dump bin_int64_bits.writer (Int64.succ (Int64.of_int max_int))
+            in
+            "Int_overflow (positive)" @? expect_read_error Int_overflow 0 (fun () ->
+              Read.bin_read_int_64bit bad_buf_max ~pos_ref:(ref 0));
+            let bad_buf_min =
+              bin_dump bin_int64_bits.writer (Int64.pred (Int64.of_int min_int))
+            in
+            "Int_overflow (negative)" @? expect_read_error Int_overflow 0 (fun () ->
+              Read.bin_read_int_64bit bad_buf_min ~pos_ref:(ref 0));
+          );
+
+        "network16_int" >::
+          (fun () ->
+            check_all 2 "network16_int"
+              Read.bin_read_network16_int
+              Write.bin_write_network16_int
+              [
+                (* No negative numbers - ambiguous on 64bit platforms *)
+                (0, "0", 2);
+                (1, "1", 2);
+              ];
+          );
+
+        "network32_int" >::
+          (fun () ->
+            check_all 4 "network32_int"
+              Read.bin_read_network32_int
+              Write.bin_write_network32_int
+              [
+                (* No negative numbers - ambiguous on 64bit platforms *)
+                (0, "0", 4);
+                (1, "1", 4);
+              ];
+          );
+
+        "network32_int32" >::
+          (fun () ->
+            check_all 4 "network32_int32"
+              Read.bin_read_network32_int32
+              Write.bin_write_network32_int32
+              [
+                (-1l, "-1", 4);
+                (0l, "0", 4);
+                (1l, "1", 4);
+              ];
+          );
+
+        "network64_int" >::
+          (fun () ->
+            check_all 8 "network64_int"
+              Read.bin_read_network64_int
+              Write.bin_write_network64_int
+              [
+                (-1, "-1", 8);
+                (0, "0", 8);
+                (1, "1", 8);
+              ];
+          );
+
+        "network64_int64" >::
+          (fun () ->
+            check_all 8 "network64_int64"
+              Read.bin_read_network64_int64
+              Write.bin_write_network64_int64
+              [
+                (-1L, "-1", 8);
+                (0L, "0", 8);
+                (1L, "1", 8);
+              ];
+          );
       ]
 end
 
@@ -852,51 +991,59 @@ module ML = Make (struct let kind = "ml" end) (Read_ml) (Write_ml)
 module C = Make (struct let kind = "c" end) (Read_c) (Write_c)
 
 module Common = struct
-  type t = A | B of string | C of t list
+  type tuple = float * string * int64
+  with bin_io
+
+  type 'a record = { a : int; b : 'a; c : 'b. 'b option }
+  with bin_io
+
+  type 'a singleton_record = { y : 'a }
+  with bin_io
+
+  type 'a sum = Foo | Bar of int | Bla of 'a * string
+  with bin_io
+
+  type 'a variant = [ `Foo | `Bar of int | `Bla of 'a * string ]
+  with bin_io
+
+  type 'a poly_app = (tuple * int singleton_record * 'a record) variant sum list
+  with bin_io
+
+  type 'a rec_t1 = RecFoo1 of 'a rec_t2
+  and 'a rec_t2 = RecFoo2 of 'a poly_app * 'a rec_t1 | RecNone
+  with bin_io
+
+  type 'a poly_id = 'a rec_t1
+  with bin_io
+
+  type el = float poly_id
+  with bin_io
+
+  type els = el array
   with bin_io
 
   let test =
     "Bin_prot_common" >:::
       [
-        "Utils.bin_dump and Utils.Read_buf.next" >::
+        "Utils.bin_dump" >::
           (fun () ->
-            let x =
-              C
-                [
-                  A;
-                  B "asdf";
-                  B "fdsa";
-                  C
-                    [
-                      A;
-                      A;
-                    ];
-                ]
+            let el =
+              let record = { a = 17; b = 2.78; c = None } in
+              let arg = (3.1, "foo", 42L), { y = 4321 }, record in
+              let variant = `Bla (arg, "fdsa") in
+              let sum = Bla (variant, "asdf") in
+              let poly_app = [ sum ] in
+              RecFoo1 (RecFoo2 (poly_app, RecFoo1 RecNone))
             in
-            let all_buf = bin_dump ~with_size:true bin_sw_arg_t x in
-            let len = Array1.dim all_buf in
-            for last = 0 to len do
-              let buf = Array1.sub all_buf 0 last in
-              let rbuf = Read_buf.create () in
-              let outer_pos_ref = ref 0 in
-              match
-                Read_buf.next bin_read_t buf ~pos_ref:outer_pos_ref rbuf
-              with
-              | None ->
-                  "outer None: position not at end" @? (!outer_pos_ref = last);
-                  let buf = Array1.sub all_buf last (len - last) in
-                  let inner_pos_ref = ref 0 in
-                  (match
-                    Read_buf.next bin_read_t buf ~pos_ref:inner_pos_ref rbuf
-                  with
-                  | None -> "inner: None" @? false
-                  | Some v ->
-                      "inner: position not at end" @? (!inner_pos_ref = len - last);
-                      "inner: unequal" @? (v = x))
-              | Some v ->
-                  "outer Some: position not at end" @? (!outer_pos_ref = last);
-                  "outer: unequal" @? (v = x)
-            done
+            let els = Array.create 10 el in
+            let buf = bin_dump ~header:true bin_els.writer els in
+
+            let pos_ref = ref 0 in
+            let els_len = Read_ml.bin_read_int_64bit buf ~pos_ref in
+            "pos_ref for length incorrect" @? (!pos_ref = 8);
+            "els_len disagrees with bin_size" @? (els_len = bin_size_els els);
+            let new_els = bin_read_els buf ~pos_ref in
+            "new_els and els not equal" @? (els = new_els)
           );
       ]
 end
